// TiniServlet.java - A demo 1-Wire servlet for TINI.
//
// Copyright (C) 1999-2002  Smart Software Consulting
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// Smart Software Consulting
// 1688 Silverwood Court
// Danville, CA  94526-3079
// USA
//
// http://www.smartsc.com
//

import java.io.IOException;
import java.io.PrintWriter;

import java.util.Date;

import javax.servlet.*;
import javax.servlet.http.*;

//import com.dalsemi.system.Clock;
import com.dalsemi.system.TINIOS;

import com.dalsemi.onewire.OneWireException;
import com.dalsemi.onewire.adapter.DSPortAdapter;
import com.dalsemi.onewire.adapter.OneWireIOException;
import com.dalsemi.onewire.adapter.TINIExternalAdapter;
import com.dalsemi.onewire.adapter.TINIInternalAdapter;
import com.dalsemi.onewire.utils.CRC8;

public class TiniServlet
extends HttpServlet
implements SingleThreadModel
{
	public void init()
	throws ServletException
	{
		super.init();
		internalAdapter = new TINIInternalAdapter();
		externalAdapter = new TINIExternalAdapter();
		String sProfile = getInitParameter( "profile");
		if( sProfile != null)
			profile = Boolean.valueOf( sProfile).booleanValue();
	}

	public void doPost( HttpServletRequest req, HttpServletResponse res)
	throws ServletException, IOException
	{
		doGet( req, res);
	}

	public void doGet( HttpServletRequest req, HttpServletResponse res)
	throws ServletException, IOException
	{
		profileEnter( "doGet");

		// Do HTTP header stuff
		res.setContentType( "text/html");

		// Get Writer
		PrintWriter pw = res.getWriter();

		// Start HTML
		pw.println( "<html>");
		pw.println( "<head><title>TINI Servlet</title></head>");
		pw.println( "<body>");

		// Send HTML body
		try { sendBody( pw); }
		catch( OneWireException ibe)
		{
			pw.println( "<H1>Got a OneWireException!</H1>");
			pw.println( ibe);
		}

		// Finish HTML
		pw.println( "</body></html>");
		profileExit( "doGet");
	}

	public void sendBody( PrintWriter pw)
	throws OneWireException
	{
		profileEnter( "sendBody");
		pw.println( "<H1>TINI Statistics</H1>");
		pw.print  ( "This page was generated by TiniServlet at <b>");
		pw.print  ( getCurrentTime());
		pw.println( "</b>.<p>");

		pw.print  ( "TINI Serial Number is <b>");
		pw.print  ( TINIOS.getTINISerialNumber());
		pw.println( "</b>.<br>");

		String hwVersion = TINIOS.getTINIHWVersion();
		if( hwVersion == null) hwVersion = "Unknown";
		pw.print  ( "TINI Hardware Version is <b>");
		pw.print  ( hwVersion);
		pw.println( "</b>.<br>");

		pw.print  ( "TINI OS Firmware Version is <b>");
		pw.print  ( TINIOS.getTINIOSFirmwareVersion());
		pw.println( "</b>.<p>");

		long uptime = TINIOS.uptimeMillis() / 1000;
		long uptimeDays = uptime / (60 * 60 * 24);
		long uptimeHours = (uptime % (60 * 60 * 24)) / (60 * 60);
		long uptimeMinutes = (uptime % (60 * 60)) / 60;
		long uptimeSeconds = uptime % 60;

		pw.print  ( "This TINI has been up for <b>");
		pw.print  ( uptime);
		pw.print  ( " seconds (");
		pw.print  ( uptimeDays);
		pw.print  ( " days, ");
		pw.print  ( uptimeHours);
		pw.print  ( " hours, ");
		pw.print  ( uptimeMinutes);
		pw.print  ( " minutes, ");
		pw.print  ( uptimeSeconds);
		pw.print  ( " seconds)</b>.<p>");

		pw.println( "<H1>Current 1-Wire Configuration</H1>");
		pw.println( "This shows the ROM IDs currently present on this TINI's ");
		pw.println( "internal and external 1-Wire interfaces.");
		pw.println( "<p>");

		pw.println( "For each device found with a family code of '09' ");
		pw.println( "(e.g. DS2502) or '89' (e.g DS2502-E48), ");
		pw.println( "this servlet shows the data ");
		pw.println( "retrieved from the four 32 byte pages of memory ");
		pw.println( "as well as the 8 status bytes.  The memory and ");
		pw.println( "status bytes are retrieved and displayed along with ");
		pw.println( "an 8-bit CRC for each 32 byte page and the status ");
		pw.println( "bytes.<p>");

		pw.println( "For each device found with a family code of '10' ");
		pw.println( "(e.g. DS1920, a 1-Wire temperature device), this ");
		pw.println( "servlet performs a temperature conversion and shows ");
		pw.println( "the data retrieved from the temperature device's ");
		pw.println( "scratch pad (including the 8-bit CRC).  From these ");
		pw.println( "data, the temperature is calculated and displayed.<p>");

		pw.println( "<H2>TINI Internal 1-Wire</H2>");
		pw.println( "The internal 1-Wire bus of TINI typically contains ");
		pw.println( "only one device, a DS2502-E48. ");
		pw.println( "The DS2502-E48 device has a Family Code of 89 (hex).<p>");
		pw.println( "The DS2502-E48 is pre-programmed with a ");
		pw.println( "UniqueWare file in Page 0. The two interesting things ");
		pw.println( "in this file are:<ol><li>The UniqueWare project ID ");
		pw.println( "(00001129), found in bytes 0004 through 0001<li>The ");
		pw.println( "ethernet address of the TINI unit that generated this ");
		pw.println( "page (found in bytes 000A through 0005).</ol><p>");

		pw.println( "<b>The Internal 1-Wire bus of this TINI unit ");
		pw.println( "contains the following ROM IDs...</b>");

		int deviceCount = dumpOneWire( pw, internalAdapter);

		pw.print( "Detected " + deviceCount + " device");
		if( deviceCount != 1) pw.print( "s");
		pw.println( ".");

		pw.println( "<H2>TINI External 1-Wire</H2>");
		pw.println( "<b>The External 1-Wire bus of this TINI unit ");
		pw.println( "contains the following ROM IDs...</b>");

		deviceCount = dumpOneWire( pw, externalAdapter);

		pw.print( "Detected " + deviceCount + " device");
		if( deviceCount != 1) pw.print( "s");
		pw.println( ".");
		profileExit( "sendBody");
	}

	public int dumpOneWire( PrintWriter pw, DSPortAdapter adapter)
	throws OneWireException
	{
		profileEnter( "dumpOneWire");
		int deviceCount = 0;
		byte[] romId = new byte[8];

		adapter.reset();
		if( adapter.findFirstDevice())
		{
			pw.println( "<UL>");

			do
			{
				++deviceCount;
				adapter.getAddress( romId);

				pw.print( "<LI>");
				pw.print( "<code>");
				pw.print( adapter.getAddressAsString());
				pw.print( "</code>");

				if( (romId[0] & 0x7f) == DS2502)
				{
					pw.print( " - Displaying memory of this ");
					if( (romId[0] & 0x80) == 0x80
					&&  ((romId[6]<<4) | (romId[5]>>4)) == UNIQUE_WARE)
					{
						pw.print( "UniqueWare ");
					}
					pw.println( "device.");

					try
					{
						dump2502( pw, adapter);
					}
					catch( OneWireIOException owioe)
					{
						pw.print( "<br>Error reading DS2502! (");
						pw.print( owioe.getMessage());
						pw.println( ")");
					}
				}
				else if( romId[0] == DS1920)
				{
					pw.println( " - Performing temperature conversion.");

					try
					{
						dump1920( pw, adapter);
					}
					catch( OneWireIOException owioe)
					{
						pw.print( "<br>Error reading DS1920! (");
						pw.print( owioe.getMessage());
						pw.println( ")");
					}
				}

			} while( adapter.findNextDevice());

			pw.println( "</UL>");
		}

		profileExit( "dumpOneWire");
		return deviceCount;
	}

	public void dump1920( PrintWriter pw, DSPortAdapter adapter)
	throws OneWireException
	{
		profileEnter( "dump1920");
		byte[] romId = new byte[8];
		adapter.getAddress( romId);
		if( romId[0] != DS1920)
		{
			return;
		}

		int command;
		byte[] scratchpadBlock = new byte[9];

		// Start conversion
		command = CONVERT_TEMP;
		adapter.select( romId);
		adapter.putByte( command);

		// Wait 0.75 seconds for it to finish
		try { Thread.sleep( 750); }
		catch( InterruptedException ie) {}

		// Prepare block for reading
		for( int i = 0; i < scratchpadBlock.length; ++i)
			scratchpadBlock[i] = (byte)0xff;

		// Read results
		adapter.select( romId);
		adapter.putByte( READ_SCRATCHPAD);
		adapter.dataBlock( scratchpadBlock, 0, scratchpadBlock.length);

		// If CRC error
		int calcCRC = CRC8.compute( scratchpadBlock, 0);
		if( calcCRC != 0)
		{
			// Reset 1-Wire bus
			adapter.reset();
			// Throw exception
			throw new OneWireIOException(
				crcErrorString( 0, calcCRC) );
		}

		// Calc temps
		int tempC2 = scratchpadBlock[0];
		int sign   = scratchpadBlock[1];
		// If the MSB of the 9-bit temperature value is set
		if(sign != 0)
		{
			// Subtract 2^8 from the lower 8-bits of the temperature value.
			tempC2 -= 0x100;
		}
		int tempF2 = tempC2 * 9 / 5 + 64;

		// Dump data
		pw.print  ( "<pre><b>");
		pw.print  ( "            CRC ");
		pw.println( "07 06 05 04 03 02 01 00</b>");

		pw.print( "<b>Scratchpad:</b>  ");
		pw.print( toHexString(
			scratchpadBlock, scratchpadBlock.length-1, 0 ));
		pw.println();

		pw.print  ( "<b>Temperature:</b> " + (tempC2>>1) + ".");
		pw.print  ( (tempC2&1)==0 ? "0" : "5");
		pw.print  ( "&deg;C (" + (tempF2>>1) + ".");
		pw.print  ( (tempF2&1)==0 ? "0" : "5");
		pw.println( "&deg;F)");
		pw.println( "</pre>");

		// Reset 1-Wire bus
		adapter.reset();
		profileExit( "dump1920");
	}

	public void dump2502( PrintWriter pw, DSPortAdapter adapter)
	throws OneWireException
	{
		profileEnter( "dump2502");
		byte[] block;
		byte[] romId = new byte[8];
		adapter.getAddress( romId);
		if( (romId[0] & 0x7f) != DS2502)
		{
			return;
		}

		pw.print  ( "<pre><b>");
		pw.print  ( "        CRC ");
		pw.print  ( "1F 1E 1D 1C 1B 1A 19 18 17 16 15 14 13 12 11 10 ");
		pw.println( "0F 0E 0D 0C 0B 0A 09 08 07 06 05 04 03 02 01 00</b>");
		for( int page = 0; page < 4; ++page)
		{
			pw.print( "<b>Page " + page + ":</b>  ");

			try
			{
				block = read2502Page( adapter, page);
				pw.println( toHexString( block, block.length-1, 0));
			}
			catch( OneWireIOException owioe)
			{
				pw.println( "</pre>");
				throw owioe;
			}
		}
		pw.println( "</pre>");

		pw.print  ( "<pre><b>");
		pw.println( "        CRC 07 06 05 04 03 02 01 00</b>");
		pw.print  ( "<b>Status:</b>  ");
		try
		{
			block = read2502Page( adapter, -1);
			pw.println( toHexString( block, block.length-1, 0));
		}
		catch( OneWireIOException owioe)
		{
			pw.println( "</pre>");
			throw owioe;
		}
		pw.println( "</pre>");

		// Reset 1-Wire bus
		adapter.reset();
		profileExit( "dump2502");
	}

	public byte[] read2502Page( DSPortAdapter adapter, int page)
	throws OneWireException
	{
		profileEnter( "read2502Page");
		// Variables for 1-wire access
		byte command;
		int address;
		byte[] dataBlock;

		// If reading status page
		if( page == -1)
		{
			command = READ_STATUS;
			address = 0;
			dataBlock = new byte[9];
		}
		else
		{
			command = READ_MEMORY_CRC;
			address = page << 5;
			dataBlock = new byte[33];
		}

		// Init dataBlock for reading data.
		for( int i = 0; i < dataBlock.length; ++i)
			dataBlock[i] = (byte)0xff;

		// Send command and address to device
		sendCommandBlock( adapter, command, address);

		// Read sub-block
		adapter.dataBlock( dataBlock, 0, dataBlock.length);

		// If CRC error
		int calcCRC = CRC8.compute( dataBlock, 0);
		if( calcCRC != 0)
		{
			// Reset 1-Wire bus
			adapter.reset();
			// Throw exception
			throw new OneWireIOException(
				crcErrorString( 0, calcCRC) );
		}

		profileExit( "read2502Page");
		return dataBlock;
	}

	public void sendCommandBlock(
		DSPortAdapter adapter, byte command, int address)
	throws OneWireException
	{
		profileEnter( "sendCommandBlock");
		byte[] commandBlock = new byte[4];

		// Init command black
		commandBlock[0] = command;
		commandBlock[1] = (byte)(address & 0xff);
		commandBlock[2] = (byte)((address >> 8) & 0xff);
		commandBlock[3] = (byte)0xff; // Will receive CRC

		adapter.select( adapter.getAddressAsLong());
		adapter.dataBlock( commandBlock, 0, commandBlock.length);

		// If CRC error
		int calcCRC = CRC8.compute( commandBlock, 0);
		if( calcCRC != 0)
		{
			// Reset 1-Wire bus
			adapter.reset();
			// Throw exception
			throw new OneWireIOException(
				crcErrorString( 0, calcCRC) );
		}
		profileExit( "sendCommandBlock");
	}

	public String crcErrorString( int expected, int got)
	{
		StringBuffer sb = new StringBuffer( "CRC error: expected ");
		if( (expected & 0xf0) == 0) sb.append( '0');
		sb.append( Integer.toHexString( expected & 0xff));
		sb.append( ", got ");
		if( (got & 0xf0) == 0) sb.append( '0');
		sb.append( Integer.toHexString( got & 0xff));
		return sb.toString();
	}

	public static String toHexString( byte[] block, int first, int last)
	{
		profileEnter( "toHexString");
		int incr = 1;
		int len = last - first + 1;
		if( first > last)
		{
			incr = -1;
			len = 2 - len;
		}

		StringBuffer sb = new StringBuffer( 3 * len);

		for( int i = first;
		     (first <= i && i <= last) || (last <= i && i <= first);
		     i += incr)
		{
			if( (block[i] & 0xf0) == 0) sb.append( "0");
			sb.append( Integer.toHexString( block[i] & 0xff));

			sb.append( " ");
		}
		profileExit( "toHexString");
		return sb.toString().trim().toUpperCase();
	}

	public static String getCurrentTime()
	{
		return new Date().toString();
	}

	public static final byte DS2502          = (byte)0x09;
	public static final byte DS1920          = (byte)0x10;
	public static final byte CONVERT_TEMP    = (byte)0x44;
	public static final byte READ_STATUS     = (byte)0xaa;
	public static final byte READ_SCRATCHPAD = (byte)0xbe;
	public static final byte READ_MEMORY_CRC = (byte)0xc3;

	public static final int UNIQUE_WARE     = 0x5e7;

	private TINIInternalAdapter internalAdapter;
	private TINIExternalAdapter externalAdapter;

	private static boolean profile = false;

	private static void profileEnter( String name)
	{
		if( profile)
			System.out.println( "Entering " + name + " " +
				String.valueOf( System.currentTimeMillis()) );
	}

	private static void profileExit( String name)
	{
		if( profile)
			System.out.println( "Exiting " + name + " " +
				String.valueOf( System.currentTimeMillis()) );
	}
}
